/*
 * LLHAL_GPIO.cpp
 *
 * Copyright 2017 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/* Includes ------------------------------------------------------------------*/

#include "hal_mbed.h"

static analogPin_t analogs[NUMBER_OF_ANALOG_PINS] = {
    {P1_8, (ADCName)-1},
    {P1_9, (ADCName)-1},
    {P1_10, (ADCName)-1},
    {P1_11, (ADCName)-1},
    {P1_13, (ADCName)-1},
    {P1_15, (ADCName)-1}
};

static pwmoutPin_t pwms[NUMBER_OF_PWM_CHANNELS] = {
    {P4_4     , (PWMName)-1},
    {P4_6     , (PWMName)-1},
    {P4_7     , (PWMName)-1},
    {P8_14    , (PWMName)-1},
    {P8_15    , (PWMName)-1},
    {P8_13    , (PWMName)-1},
    {P8_11    , (PWMName)-1},
    {P4_5     , (PWMName)-1}
};
/* Macros --------------------------------------------------------------------*/
#define __hal_get_gpio_data(port, pin, gpio)   do { \
		if (get_gpio_data(port, pin, gpio) == HAL_GPIO_NOK) \
		return HAL_GPIO_NOK; /* invalid port or pin */ \
		} while(0)

/* Fields --------------------------------------------------------------------*/
static SingletonPtr<PlatformMutex> _mutex_init;


/* Private functions ---------------------------------------------------------*/

static int32_t get_gpio_data(int32_t port, int32_t pin, gpio_t* gpio)
{

	if (port == HAL_GPIO_PORT_MCU)
		{
			/*
			 * MCU virtual port: convert port and pin using real MCU port and pin
			 *
			 * MCU port = (pin / 16) + 1
			 * MCU pin = pin % 16
			 */

			if (pin < 0)
			{
				return HAL_GPIO_NOK;	// invalid pin
			}

			int32_t adjusted_pin=pin;
				if (pin >74){
					// PORT_5 contains 11 pins
					adjusted_pin+=5;
					if(pin>130){
						// PORT_9 contains 8 pins
						adjusted_pin+=8;
					}
				}

			return hal_get_gpio_data((adjusted_pin >> 4) + 1, adjusted_pin & 0xf, gpio);
		}
	else
	{
		return hal_get_gpio_data(port,  pin, gpio);
	}
}

static analogin_t* getAnalogInPinStrucCtrl(PinName pin)
{
	analogin_t* ret = NULL;

	for(int i=0; i<NUMBER_OF_ANALOG_PINS; i++ )
	{
		if(analogs[i].pinName == pin)
			ret = &(analogs[i].analog);
	}

	return ret;
}

static pwmout_t* getAnalogOutPinStrucCtrl(PinName pin)
{
	pwmout_t* ret = NULL;

	for(int i=0; i<NUMBER_OF_PWM_CHANNELS; i++ )
	{
		if(pwms[i].pinName == pin)
			ret = &(pwms[i]._pwm);
	}

	return ret;
}

/* API -----------------------------------------------------------------------*/
int32_t LLHAL_GPIO_IMPL_setPinMode(int32_t port, int32_t pin, int32_t mode)
{
	gpio_t gpio;
	__hal_get_gpio_data(port, pin, &gpio);

	switch (mode)
	{

		case LLHAL_GPIO_MODE_DIGITAL_INPUT:
		{
			gpio_init_inout(&gpio, gpio.pin, PIN_INPUT, PullDefault, 0);
			break;
		}
		case LLHAL_GPIO_MODE_DIGITAL_INPUT_PULLUP:
		{
			gpio_init_inout(&gpio, gpio.pin, PIN_INPUT, PullDefault, 0);
			break;
		}
		case LLHAL_GPIO_MODE_DIGITAL_OUTPUT:
		{
			gpio_init_inout(&gpio, gpio.pin, PIN_OUTPUT, PullNone, 0);
			break;
		}
		case LLHAL_GPIO_MODE_ANALOG_INPUT:
		{
			int val=0;

			analogin_t* _adc = getAnalogInPinStrucCtrl(gpio.pin);

			if (_adc != NULL)
			{
				// initalisation
				_mutex_init->lock();
				analogin_init(_adc, gpio.pin);
				_mutex_init->unlock();
			}
			else
				return HAL_GPIO_NOK;
			break;
		}
		case LLHAL_GPIO_MODE_ANALOG_OUTPUT:
		{
			pwmout_t* _pwm = getAnalogOutPinStrucCtrl(gpio.pin);

			if (_pwm != NULL)
			{
				core_util_critical_section_enter();
				pwmout_init(_pwm, gpio.pin);
				pwmout_period_ms(_pwm,2);
				core_util_critical_section_exit();
			}
			else
				return HAL_GPIO_NOK;
			break;
		}
	}

	// no error
	return HAL_GPIO_OK;
}


int32_t LLHAL_GPIO_IMPL_getDigitalValue(int32_t port, int32_t pin)
{
	// retrieve GPIO data
	gpio_t gpio;
	__hal_get_gpio_data(port, pin, &gpio);
	 gpio_init(&gpio, gpio.pin);

	return (gpio_read(&gpio)== 1 ? LLHAL_GPIO_DIGITAL_HIGH : LLHAL_GPIO_DIGITAL_LOW);
}

int32_t LLHAL_GPIO_IMPL_setDigitalValue(int32_t port, int32_t pin, int32_t value){

	// retrieve GPIO data
	gpio_t gpio;
	__hal_get_gpio_data(port, pin, &gpio);
	gpio_init(&gpio, gpio.pin);

	gpio_write(&gpio, value==LLHAL_GPIO_DIGITAL_LOW ? LLHAL_GPIO_DIGITAL_LOW: LLHAL_GPIO_DIGITAL_HIGH);
	return HAL_GPIO_OK;
}

int32_t LLHAL_GPIO_IMPL_getAnalogValue(int32_t port, int32_t pin) {

	gpio_t gpio;
	__hal_get_gpio_data(port, pin, &gpio);

	analogin_t* _adc = getAnalogInPinStrucCtrl(gpio.pin);
	if (_adc != NULL)
	{
		 _mutex_init->lock();
		 uint16_t ret = analogin_read_u16(_adc);
		 _mutex_init->unlock();

		 //Scale the ADC output on 16 bit to 12 bit value
		 ret = (ret * 0x0FFF ) / 0xFFFF;

		 return (int32_t)ret;
	}
	else
		return HAL_GPIO_NOK;
}

int32_t LLHAL_GPIO_IMPL_setAnalogValue(int32_t port, int32_t pin, int32_t percentage)
{
	// retrieve GPIO data
	gpio_t gpio;
	__hal_get_gpio_data(port, pin, &gpio);

	pwmout_t* _pwm = getAnalogOutPinStrucCtrl(gpio.pin);

	if (_pwm != NULL)
	{
		core_util_critical_section_enter();
		pwmout_write(_pwm, ((float)percentage/100));
		core_util_critical_section_exit();
		return HAL_GPIO_OK;
	}
	else
		return HAL_GPIO_NOK;
}


/* EOF -----------------------------------------------------------------------*/




