/*
 * C
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <LLNET_Util.h>
#include <stdio.h>
#include <string.h>
#include "cmsis_os.h"

#define MAC_ADDR_SEPARATOR	':'
#define MAC_ADDR_STR_SIZE   17

/* global variable to store errno that can differ for JVM or dispatch threads */
osMutexId errno_mutex;
osMutexDef(errno_mutex);
uint8_t errno_mutex_init = 0;
static errno_thread_t jvm_errno = {0};
static errno_thread_t dispatch_errno = {0};

static int32_t convertHexStringToInt(uint8_t* str, int32_t offset, int32_t length);

uint32_t computeCRC24(const char* buffer, int32_t length) {
	uint32_t result = 0;
    uint32_t polynom = 0x005A6000; // BTLE polynom
	uint32_t i,j;
    for(i = 0; i < length; ++i) {
        char b = buffer[i];
        for (j = 0; j < 8; ++j) {
            int bitSet = (result ^ b) & 1;
            result = result >> 1;
            b = b >> 1;
            if (bitSet) {
                result |= 1 << 23;
                result = result ^ polynom;
            }
        }
    }
    return result;
}

/**
 * Parse a MAC address in string format and return the associated byte array.
 * Return an array of 0 if an error occurred.
 * @param str MAC address in string format
 * @param dst destination MAC address array pointer
 * @param dstOffset destination MAC address array pointer offset
 * @param dstLength destination MAC address array length
 */
void parseMacAddr(uint8_t* str, uint8_t* dst, int32_t dstOffset, int32_t dstLength){
	uint32_t count = 0;
	int32_t dstIndex = 0;
	int32_t strLen = strlen((const char*)str);
	if(strLen != MAC_ADDR_STR_SIZE){
		return;
	}
	for(int32_t i = -1; ++i < strLen;){
		if(str[i] == MAC_ADDR_SEPARATOR){
			if((count == 2) && (dstIndex < (dstLength-1))){
				dst[(dstIndex++) + dstOffset] = (uint8_t)convertHexStringToInt(&str[i-2], 0, 2);
				count = 0;
			}else{
				// parse error
				for(int32_t j = -1; ++j < dstLength;){
					dst[dstOffset + j] = 0;
				}
				return;
			}
		}else{
			++count;
		}
	}
	// last byte
	dst[dstIndex + dstOffset] = (uint8_t)convertHexStringToInt(&str[strLen-2], 0, 2);
}

/**
 * Convert an hexadecimal string (without '0x' prefix) to an integer
 * @param str a hexadecimal string
 * @param offset hexadecimal string offset
 * @param length hexadecimal string length
 * @return integer conversion of the hexadecimal string
 */
static int32_t convertHexStringToInt(uint8_t* str, int32_t offset, int32_t length){
	int32_t res = 0;
	for(int32_t i = -1; ++i < length;){
		int32_t tmp = str[offset + i] - '0';
		if(tmp > 10){
			tmp = str[offset + i] - 'A';
			if(tmp > 10){
				tmp = str[offset + i] - 'a';
			}
			tmp += 10;
		}
		res += tmp  << (4 * (length - 1 - i));
	}
	return res;
}

static void LLNET_Util_init_errno_mutex(void)
{
	errno_mutex = osMutexCreate(osMutex(errno_mutex));
	if (errno_mutex == NULL)
		printf("BSP ERROR : errno_mutex create error\n");
	errno_mutex_init = 1;
}

void LLNET_Util_init_errno_jmv(void)
{
	if (!errno_mutex_init)
		LLNET_Util_init_errno_mutex();

	jvm_errno.threadId = (int32_t)(osThreadGetId());
}

void LLNET_Util_init_errno_dispatch(void)
{
	if (!errno_mutex_init)
		LLNET_Util_init_errno_mutex();

	dispatch_errno.threadId = (int32_t)(osThreadGetId());
}

void LLNET_Util_set_errno(int32_t err)
{
	// Lock errno mutex
	if (osMutexWait(errno_mutex, osWaitForever) != osOK)
		printf("BSP ERROR : errno_mutex Lock error\n");

	if (jvm_errno.threadId != 0 && dispatch_errno.threadId != 0)
	{
		int32_t currentThread = (int32_t)(osThreadGetId());
		if (currentThread == jvm_errno.threadId)
			jvm_errno.errnoVal = err;
		else if (currentThread == dispatch_errno.threadId)
			dispatch_errno.errnoVal = err;
		else
			printf("BSP ERROR : thread ID mismatch for setting errno\n");
	}
	else
		printf("BSP ERROR : thread ID not initialized for setting errno\n");

	// Unlock errno mutex
	if (osMutexRelease(errno_mutex) != osOK)
	    	printf("BSP ERROR : errno_mutex UnLock error\n");
}

int32_t LLNET_Util_get_errno(void)
{
	// Lock errno mutex
	if (osMutexWait(errno_mutex, osWaitForever) != osOK)
		printf("BSP ERROR : errno_mutex Lock error\n");

	int32_t errno = 0;
	if (jvm_errno.threadId != 0 && dispatch_errno.threadId != 0)
	{
		int32_t currentThread = (int32_t)(osThreadGetId());
		if (currentThread == jvm_errno.threadId)
			errno = jvm_errno.errnoVal;
		else if (currentThread == dispatch_errno.threadId)
			errno = dispatch_errno.errnoVal;
		else
			printf("BSP ERROR : thread ID mismatch for setting errno\n");
	}
	else
		printf("BSP ERROR : thread ID not initialized for getting errno\n");

	// Unlock errno mutex
	if (osMutexRelease(errno_mutex) != osOK)
	    	printf("BSP ERROR : errno_mutex UnLock error\n");

	return errno;
}

/**
 * Max number of JVM Threads
 */
uint32_t LLNET_Util_getNaxJVMThread(void)
{
	return (uint32_t)&_java_max_nb_standaloneEngines;
}
