/*
 * C
 *
 * Copyright 2015-2017 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
#include <LLNET_SSL_CONTEXT_impl.h>
#include <LLNET_SSL_CONSTANTS.h>
#include <LLNET_SSL_ERRORS.h>
#include <LLNET_SSL_MEM.h>
#include <LLNET_SSL_utils_mbedtls.h>
#include "LLNET_SSL_verifyCallback.h"
#include <stdlib.h>

#include "mbedtls/debug.h"
#include "mbedtls/net.h"
#include "mbedtls/ssl.h"
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#include "mbedtls/error.h"

/* Check that the custom Mbedtls configuration file (MBEDTLS_CONFIG_FILE) is defined by the build definition*/
#ifndef MBEDTLS_CONFIG_FILE
#error "Custom configuration file (MBEDTLS_CONFIG_FILE) should be defined by the build definition"
#endif

#ifdef __cplusplus
	extern "C" {
#endif

/* ----------- external function and variables ----------- */
extern int32_t LLNET_SSL_TranslateReturnCode(int32_t wolfSSL_error);
extern mbedtls_ctr_drbg_context ctr_drbg;

/* ----------- Private API  -----------*/
#if MBEDTLS_DEBUG_LEVEL > 0
/**
 * Debug callback for mbed TLS
 * Just prints on the USB serial port
 */
void my_debug(void *ctx, int level, const char *file, int line,
					 const char *str)
{
	const char *p, *basename;
	(void) ctx;

	/* Extract basename from file */
	for(p = basename = file; *p != '\0'; p++) {
		if(*p == '/' || *p == '\\') {
			basename = p + 1;
		}
	}

	printf("%s:%04d: |%d| %s", basename, line, level, str);
}
#endif

/* ----------- API  -----------*/

int32_t LLNET_SSL_CONTEXT_IMPL_createClientContext(int32_t protocol, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(method=%d)\n", __func__, protocol);

	mbedtls_ssl_config* conf = (mbedtls_ssl_config*)LLNET_SSL_MEM_allocate(sizeof(mbedtls_ssl_config));

	if (conf != NULL)
	{
		mbedtls_ssl_config_init(conf);

		int ret;
		if ((ret = mbedtls_ssl_config_defaults(conf,
						MBEDTLS_SSL_IS_CLIENT,
						MBEDTLS_SSL_TRANSPORT_STREAM,
						MBEDTLS_SSL_PRESET_DEFAULT)) != 0) {
			LLNET_SSL_DEBUG_MBEDTLS_TRACE("mbedtls_ssl_config_defaults", ret);
		}

		switch (protocol) {

			case SSLv3_PROTOCOL:
				mbedtls_ssl_conf_min_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_0 );
				mbedtls_ssl_conf_max_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_0 );
				break;
			case TLSv1_PROTOCOL:
				mbedtls_ssl_conf_min_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_1 );
				mbedtls_ssl_conf_max_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_1 );
				break;
			case TLSv1_1_PROTOCOL:
				mbedtls_ssl_conf_min_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_2 );
				mbedtls_ssl_conf_max_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_2 );
				break;
			case TLSv1_2_PROTOCOL:
				mbedtls_ssl_conf_min_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_3 );
				mbedtls_ssl_conf_max_version(conf, MBEDTLS_SSL_MAJOR_VERSION_3, MBEDTLS_SSL_MINOR_VERSION_3 );
				break;
			default:
				break;
		}

		mbedtls_ssl_conf_rng(conf, mbedtls_ctr_drbg_random, &ctr_drbg);

		/* It is possible to disable authentication by passing
		 * MBEDTLS_SSL_VERIFY_NONE in the call to mbedtls_ssl_conf_authmode()
		 */
		mbedtls_ssl_conf_authmode(conf, MBEDTLS_SSL_VERIFY_REQUIRED);

		/*
		 * Change certificate verification profile, allocate and initialize new profile
		 */
		mbedtls_x509_crt_profile* crt_profile = (mbedtls_x509_crt_profile*)LLNET_SSL_MEM_allocate(sizeof(mbedtls_x509_crt_profile));
		if (crt_profile == NULL)
			return J_CREATE_SSL_CONTEXT_ERROR;
		/* Hashes from MD5 and above */
		crt_profile->allowed_mds = MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_MD5 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_SHA1 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_RIPEMD160 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_SHA224 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_SHA256 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_SHA384 ) |
								   MBEDTLS_X509_ID_FLAG( MBEDTLS_MD_SHA512 );
		/* Any PK alg    */
		crt_profile->allowed_pks = 0xFFFFFFF;
		/* Any curve     */
		crt_profile->allowed_curves = 0xFFFFFFF;
		/* RSA min lenght */
		crt_profile->rsa_min_bitlen = 1024;

		mbedtls_ssl_conf_cert_profile(conf, crt_profile);

		/* Allocate and initialize certificate verify context*/
		cert_verify_ctx* verify_ctx = (cert_verify_ctx*)LLNET_SSL_MEM_allocate(sizeof(cert_verify_ctx));
		if (verify_ctx == NULL)
			return J_CREATE_SSL_CONTEXT_ERROR;

        verify_ctx->conf = conf;
        verify_ctx->isUnTrustCA = 0;

		mbedtls_ssl_conf_verify(conf, LLNET_SSL_VERIFY_verifyCallback, (void*)verify_ctx);

#if MBEDTLS_DEBUG_LEVEL > 0
			mbedtls_ssl_conf_dbg(conf, my_debug, NULL);
			mbedtls_debug_set_threshold(MBEDTLS_DEBUG_LEVEL);
#endif

		LLNET_SSL_DEBUG_TRACE("%s(method=%d) return ctx=%p\n", __func__, protocol,conf);

		return (int32_t)conf;
	}

	return J_CREATE_SSL_CONTEXT_ERROR; //error
}

int32_t LLNET_SSL_CONTEXT_IMPL_addTrustedCert(int32_t context, int8_t *cert, int32_t off, int32_t len, int32_t format, uint8_t retry){

	LLNET_SSL_DEBUG_TRACE("%s(context=%p, cert=%p, off=%d, len=%d, format=%d)\n", __func__, context, cert, cert, len, format);

	int32_t ret = J_SSL_NO_ERROR;

	mbedtls_ssl_config* conf = (mbedtls_ssl_config*)(context);
	mbedtls_x509_crt* cacert = NULL;
	if (conf->ca_chain == NULL)
	{
		cacert = (mbedtls_x509_crt*)LLNET_SSL_MEM_allocate(sizeof(mbedtls_x509_crt));
		mbedtls_x509_crt_init(cacert);
	}
	else
	{
		cacert = conf->ca_chain;
	}

	if (conf != NULL && cacert != NULL)
	{
		int ret;
		if ((ret = mbedtls_x509_crt_parse(cacert, (const unsigned char *) (cert+off), len)) != 0) {
			LLNET_SSL_DEBUG_MBEDTLS_TRACE("mbedtls_x509_crt_parse", ret);
			return LLNET_SSL_TranslateReturnCode(ret);
		}

		if (conf->ca_chain == NULL){
			mbedtls_ssl_conf_ca_chain(conf, cacert, NULL);
		}
	}
	else
	{
		ret = J_CREATE_SSL_CONTEXT_ERROR;
	}

    return ret;
}


int32_t LLNET_SSL_CONTEXT_IMPL_clearTrustStore(int32_t context, uint8_t retry){

	LLNET_SSL_DEBUG_TRACE("%s(context=%p)\n", __func__, context);
	mbedtls_ssl_config* conf = (mbedtls_ssl_config*)(context);

	if (conf != NULL)
	{
		if (conf->ca_chain != NULL)
		{
			void* chain_ptr = (void*)conf->ca_chain;
			mbedtls_ssl_conf_ca_chain(conf, NULL, NULL);
			mbedtls_x509_crt_free(chain_ptr);
			LLNET_SSL_MEM_free(chain_ptr);
		}
	}

	return J_SSL_NO_ERROR;
}

int32_t LLNET_SSL_CONTEXT_IMPL_closeContext(int32_t contextID, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(context=%p)\n", __func__, contextID);
	mbedtls_ssl_config* conf = (mbedtls_ssl_config*)(contextID);

	if (conf != NULL)
	{
		if (conf->ca_chain != NULL)
		{
			void* chain_ptr = (void*)conf->ca_chain;
			mbedtls_ssl_conf_ca_chain(conf, NULL, NULL);
			mbedtls_x509_crt_free(chain_ptr);
			LLNET_SSL_MEM_free(chain_ptr);
		}

		if (conf->p_vrfy != NULL)
		{
			void* vrfy_ptr = (void*)conf->p_vrfy;
			mbedtls_ssl_conf_verify(conf, NULL, NULL);
			LLNET_SSL_MEM_free(vrfy_ptr);
		}

		if (conf->cert_profile != NULL)
		{
			void* profile_ptr = (void*)conf->cert_profile;
			mbedtls_ssl_conf_cert_profile(conf, NULL);
			LLNET_SSL_MEM_free(profile_ptr);
		}

		mbedtls_ssl_config_free(conf);
		LLNET_SSL_MEM_free((void*)conf);
	}

	return J_SSL_NO_ERROR;
}
#ifdef __cplusplus
	}
#endif
