/*
 * C
 *
 * Copyright 2015-2017 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * This file converts the input events in MicroUI Event Generator events.
 */

/* Includes ------------------------------------------------------------------*/

#include "LLINPUT.h"
#include "buttons_helper_configuration.h"

// this h file is created by buildSystemMicroUI step
#include "microui_constants.h"

/* Defines -------------------------------------------------------------------*/

/*
 * Defines the three modes of buttons actions
 * When the buttons is pressed/released...
 *
 * - SEND_BUTTON_EVENT: a MicroUI Buttons event is send (press / repeat / release)
 * - SEND_COMMAND_EVENT: a MicroUI Command event is send (press / repeat only )
 * - SEND_HOME_EVENT: a MicroUI Home event is send (press only)
 *
 * The Buttons and Command events are sent to the MicroUI application.
 * The Home event is reserved to the internal framework.
 */
#define SEND_BUTTON_EVENT	(0)
#define SEND_COMMAND_EVENT	(1)
#define SEND_HOME_EVENT		(2)

/*
 * Choose the button mode selecting a value in previous list
 */
#define S1_SEND_EVENT 	    SEND_BUTTON_EVENT

/*
 * Specify which MicroUI Command to send when the mode "SEND_COMMAND_EVENT" is selected
 */
#define S1_COMMAND_EVENT	LLINPUT_COMMAND_ESC


/*
 * Specify the value of MicroUI Home event to send when the mode "WAKEUP_SEND_HOME_EVENT" is selected
 */
#define HOME (1)

/* Private functions ---------------------------------------------------------*/

/*
 * Get MicroUI button attributes according given button ID.
 * There are two MicroUI groups of buttons in this platform. They have each an
 * unique event generator ID: MICROUI_EVENTGEN_BUTTONS.
 * Each group has n buttons; the buttons ID are 0 to n-1.
 */
static void EVENT_GENERATOR_get_button_attributes(int32_t *button_id, int32_t *send_mode, int32_t *button_event_generator_id, int32_t *command)
{
	switch (*button_id) {

	case BUTTON_S1_ID:
		*button_id = 0;
		*send_mode = S1_SEND_EVENT;
		*button_event_generator_id = MICROUI_EVENTGEN_BUTTONS;
		*command = S1_COMMAND_EVENT;
		break;

	default:
		*send_mode = -1;	// invalid mode
		break;
	}
}

/* Buttons -------------------------------------------------------------------*/

int32_t EVENT_GENERATOR_button_pressed(int32_t buttonId)
{
	int32_t send_mode, button_event_generator_id, command;
	EVENT_GENERATOR_get_button_attributes(&buttonId, &send_mode, &button_event_generator_id, &command);

	switch(send_mode)
	{
	case SEND_BUTTON_EVENT:
		// send a MicroUI Buttons "press" event
		return LLINPUT_sendButtonPressedEvent(button_event_generator_id, buttonId);

	}

	// the event has not been managed
	return LLINPUT_NOK;
}

int32_t EVENT_GENERATOR_button_repeated(int32_t buttonId)
{
	int32_t send_mode, button_event_generator_id, command;
	EVENT_GENERATOR_get_button_attributes(&buttonId, &send_mode, &button_event_generator_id, &command);

	switch(send_mode)
	{
	case SEND_BUTTON_EVENT:
		// send a MicroUI Buttons "repeat" event
		return LLINPUT_sendButtonRepeatedEvent(button_event_generator_id, buttonId);

	case SEND_COMMAND_EVENT:
		// send a MicroUI Command event
		return LLINPUT_NOK;

	case SEND_HOME_EVENT:
		// do not send a Home event on repeat event
		return LLINPUT_NOK; // the event has been managed
	}

	// the event has not been managed
	return LLINPUT_NOK;
}

int32_t EVENT_GENERATOR_button_released(int32_t buttonId)
{
	int32_t send_mode, button_event_generator_id, command;
	EVENT_GENERATOR_get_button_attributes(&buttonId, &send_mode, &button_event_generator_id, &command);

	switch(send_mode)
	{
	case SEND_BUTTON_EVENT:
		// send a MicroUI Buttons "release" event
		return LLINPUT_sendButtonReleasedEvent(button_event_generator_id, buttonId);

	case SEND_COMMAND_EVENT:
	case SEND_HOME_EVENT:
		// do not send a Command or Home event on release event
		return LLINPUT_OK; // the event has been managed
	}

	// the event has not been managed
	return LLINPUT_NOK;
}

